#define uint8_t unsigned char

#define ALU 0x80
#define PLUS 0x0
#define MIN  0x1
#define MUL  0x2
#define DIV  0x3
#define MOD  0x4
#define LT   0x5
#define GT   0x6
#define EQ   0x7

#define MEM 0x40
#define SET  0x0
#define GET  0x1
#define IMM  0x2

#define JMP 0x20
#define JT  0x0
#define JF  0x1
#define JA  0x2

uint8_t mem[32];
uint8_t st[16];
uint8_t sp;

#define push( a ) st[++sp&0x0F] = a
#define pop() st[((--sp)+1)&0x0F]

void exec( unsigned char *ip )  {
  register int a;
  register int w;

  while( 1 )  {
    switch( *ip )  {
    case JMP|JT:
        a = pop();
        if( a ) ip += *(ip+1);
        break;
    case JMP|JF:
        a = pop();
        if( !a ) ip += *(ip+1);
        break;
    case JMP|JA:
        a = pop();
        ip += *(ip+1);
        break;
    case MEM|IMM:
        ip++;
        push( *ip );
        break;
    case MEM|GET:
        a = pop();
        push(mem[(uint8_t)a]);
        break;
    case MEM|SET:
        a = pop();
        w = pop();
        mem[(uint8_t)w] = a;
        break;
    case ALU|LT:
        a = pop();
        w = pop();
        push( w < a );
        break;
    case ALU|GT:
        a = pop();
        w = pop();
        push( w > a );
        break;
    case ALU|PLUS:
        a = pop();
        w = pop();
        push( a + w );
        break;
    case ALU|EQ:
        a = pop();
        w = pop();
        push( w == a );
        break;
    case ALU|MIN:
        a = pop();
        w = pop();
        push( w - a );
        break;
    case ALU|MUL:
        a = pop();
        w = pop();
        push( a * w );
        break;
    case ALU|DIV:
        a = pop();
        w = pop();
        push( w / a );
        break;
    case ALU|MOD:
        a = pop();
        w = pop();
        push( w % a );
        break;
  	}
  }
}

